/* PicProg - A library for programming the PIC 12F629 via the
*            parallel port under Linux.
* Copyright (C) 2005 by Dirk Baechle (dl9obn@darc.de)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the 
*
* Free Software Foundation, Inc.
* 675 Mass Ave
* Cambridge
* MA 02139
* USA
*
*/


/** \file picprog.c
* Library for programming a PIC 12F629 via the parallel port under Linux.
* \author Dirk Baechle, dl9obn@darc.de
* \version 1.0
* \date 2005-10-29
*/

/* This file was created automatically from the file `picprog.nw' by NOWEB.
* If you want to make changes, please edit the source `picprog.nw'.
* A precompiled documentation can be found in `picprog.pdf' and `picprog.ps',
* respectively.
* Read it to understand why things are as they are. Thank you!
*/


#include <stdio.h>

#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <string.h>


#include "ppprog.h"
#include "picprog.h"


#include "alarm.h"


static const int DATA_IN = 0x1;
static const int CLOCK_IN = 0x2;
static const int DATA_ENABLE = 0x4;
static const int VPROG = 0x08;
static const int VDD = 0x10;
static const int CLOCK_ENABLE = 0x20;
static const int DATA_OUT = 0x4000;


/** Delay after raising Vprog */
static const int VPROG_DELAY = 100;
/** Delay after raising Vdd */
static const int VDD_DELAY = 100000;


/** Clock delay */
static const int CLOCK_DELAY = 100;

/** Available command words */
static const int LOAD_CONF = 0x00;
static const int LOAD_DATA_PM = 0x02;
static const int LOAD_DATA_DM = 0x03;
static const int READ_DATA_PM = 0x04;
static const int READ_DATA_DM = 0x05;
static const int INCREMENT_ADDRESS = 0x06;
static const int BEGIN_PROG = 0x08;
static const int BEGIN_PROG_ET = 0x18;
static const int END_PROG = 0x0A;
static const int ERASE_PM = 0x09;
static const int ERASE_DM = 0x0B;


#define PROGRAM_MAX_TRY 3


/** Maximum length of an input line for INHEX16 files */
#define MAX_INPUT 1000

/** The device file */
static int device_file = 0;


/** Tries to open the device file and stores its handle
* in the global variable \a device_file.
* @return OK for success, ERROR for an error
*/
int initLib()
{
  /* Try to open device file */
  device_file = open("/dev/ppprog", O_RDONLY);
  if (device_file < 0)
  {
    device_file = 0;
    return ERROR;
  }

  return OK;
}


/** Closes the device file.
*/
void closeLib()
{
  if (0 == device_file)
    return;

  close(device_file);
}


/** Switches off all voltages.
*/
void setInactiveMode()
{
  int data[3];

  /* Switch everything off */
  data[0] = 1;
  data[1] = 0;
  data[2] = VDD_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);

}


/** Sets programmer to programming mode.
*/
void setProgrammingMode()
{
  int data[5];

  /* Switch on Vprog first */
  data[0] = 2;
  data[1] = VPROG;
  data[2] = VPROG_DELAY;
  /* and let Vdd follow */
  data[3] = VPROG | VDD;
  data[4] = VDD_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);

}


void writeCommandWord(int command)
{
  int data[50];
  int i;

  data[0] = 12;

  for (i = 0; i < 6; ++i)
  {
    /* Raise Clock and apply data */
    data[i*4+1] = (command & 0x1) | 
                  CLOCK_IN | 
                  VPROG | 
                  VDD;
    /* Clock delay */
    data[i*4+2] = CLOCK_DELAY;

    /* Lower Clock while keeping data applied */
    data[i*4+3] = (command & 0x1) | 
                  VPROG | 
                  VDD;
    /* Clock delay */
    data[i*4+4] = CLOCK_DELAY;


    command = command >> 1;
  }

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);
}


void writeDataWord(int command)
{
  int data[150];
  int i;

  data[0] = 32;

  command = (command & 0x3FFF) << 1;
  for (i = 0; i < 16; ++i)
  {
    /* Raise Clock and apply data */
    data[i*4+1] = (command & 0x1) | 
                  CLOCK_IN | 
                  VPROG | 
                  VDD;
    /* Clock delay */
    data[i*4+2] = CLOCK_DELAY;

    /* Lower Clock while keeping data applied */
    data[i*4+3] = (command & 0x1) | 
                  VPROG | 
                  VDD;
    /* Clock delay */
    data[i*4+4] = CLOCK_DELAY;


    command = command >> 1;
  }

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);
}


int readDataWordPM()
{
  int data[5];
  int din[2];
  int word = 0;
  int mask = 0x1;
  int i;

  writeCommandWord(READ_DATA_PM);

  data[0] = 1;

  /* Disable data line */
  data[1] = DATA_ENABLE | 
            VPROG | 
            VDD;
  /* Clock delay */
  data[2] = CLOCK_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);


  for (i = 0; i < 16; ++i)
  {
    /* Raise Clock */
    data[1] = CLOCK_IN |
              DATA_ENABLE | 
              VPROG | 
              VDD;
    /* Clock delay */
    data[2] = CLOCK_DELAY;

    /* Send command sequence */
    ioctl(device_file, IOCTL_PPPROG_SEND, data);

    /* Read data bit */
    ioctl(device_file, IOCTL_PPPROG_CHECK, din);

    if ((din[0] & DATA_OUT) == DATA_OUT)
      word |= mask;

    mask = mask << 1;

    /* Lower Clock */
    data[1] = DATA_ENABLE | 
              VPROG | 
              VDD;
    /* Clock delay */
    data[2] = CLOCK_DELAY;

    /* Send command sequence */
    ioctl(device_file, IOCTL_PPPROG_SEND, data);

  }

  /* Enable data line */
  data[1] = VPROG | 
            VDD;
  /* Clock delay */
  data[2] = CLOCK_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);


  word = (word >> 1) & 0x3FFF;
  return word;
}


int readDataWordDM()
{
  int data[5];
  int din[2];
  int word = 0;
  int mask = 0x1;
  int i;

  writeCommandWord(READ_DATA_DM);

  data[0] = 1;

  /* Disable data line */
  data[1] = DATA_ENABLE | 
            VPROG | 
            VDD;
  /* Clock delay */
  data[2] = CLOCK_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);


  for (i = 0; i < 16; ++i)
  {
    /* Raise Clock */
    data[1] = CLOCK_IN |
              DATA_ENABLE | 
              VPROG | 
              VDD;
    /* Clock delay */
    data[2] = CLOCK_DELAY;

    /* Send command sequence */
    ioctl(device_file, IOCTL_PPPROG_SEND, data);

    /* Read data bit */
    ioctl(device_file, IOCTL_PPPROG_CHECK, din);

    if ((din[0] & DATA_OUT) == DATA_OUT)
      word |= mask;

    mask = mask << 1;

    /* Lower Clock */
    data[1] = DATA_ENABLE | 
              VPROG | 
              VDD;
    /* Clock delay */
    data[2] = CLOCK_DELAY;

    /* Send command sequence */
    ioctl(device_file, IOCTL_PPPROG_SEND, data);

  }

  /* Enable data line */
  data[1] = VPROG | 
            VDD;
  /* Clock delay */
  data[2] = CLOCK_DELAY;

  /* Send command sequence */
  ioctl(device_file, IOCTL_PPPROG_SEND, data);


  word = (word >> 1) & 0x3FFF;
  return word;
}


void incrementAddress()
{
  writeCommandWord(INCREMENT_ADDRESS);
}


void bulkErase()
{
  int i = 0;
  writeCommandWord(LOAD_CONF);
  writeDataWord(0x3FFF);
  for (; i < 7; ++i)
    incrementAddress();  
  writeCommandWord(ERASE_PM);
  AlarmSet(9);
  AlarmWait();
  writeCommandWord(ERASE_DM);
  AlarmSet(9);
  AlarmWait();
}


void loadDataWordPM(int data)
{
  writeCommandWord(LOAD_DATA_PM);
  writeDataWord(data);
}

void loadDataWordDM(int data)
{
  writeCommandWord(LOAD_DATA_DM);
  writeDataWord(data);
}


void loadConfig()
{
  writeCommandWord(LOAD_CONF);
  writeDataWord(0x3FFF);
  AlarmSet(9);
  AlarmWait();
}


void programInternallyTimed()
{
  writeCommandWord(BEGIN_PROG);
  AlarmSet(150);
  AlarmWait();
}


int programAndVerify(int data)
{
  int check = 0;
  int try = 0;
  for (; try < PROGRAM_MAX_TRY; ++try)
  {
    loadDataWordPM(data);
    programInternallyTimed();

    check = readDataWordPM();
    if (check == data)
    {
      AlarmSet(150);
      AlarmWait();
      return OK;
    }

  }
 
  return ERROR;
}


int programInhex16File(const char *fileName)
{
  int status = OK;
  int picAddress = 0;
  int osccal = 0;
  int configword = 0;

  FILE *f;

  char input[MAX_INPUT+2];

  int ivalue = 0;

  char *i_ptr;
  int word = 0;
  int lineaddress = 0;
  int data = 0;


  
  f = fopen(fileName, "r");
  if (NULL == f)
  {
    printf("Error: Could not open INHEX16 file %s for reading!\n", fileName);  
    return ERROR;
  }


  
  setInactiveMode();
  printf("Attach adapter to parallel port and press RETURN.\n");
  getchar();
  setProgrammingMode();
  printf("Programming mode is ON now!\n");
  printf("Press RETURN again to start.\n");
  getchar();

  
  for (; picAddress < 0x3FF; ++picAddress)
    incrementAddress();

  osccal = readDataWordPM();

  
  loadConfig();

  incrementAddress();
  incrementAddress();
  incrementAddress();
  incrementAddress();
  incrementAddress();
  incrementAddress();
  incrementAddress();
  configword = readDataWordPM();

  bulkErase();
  setInactiveMode();
  setProgrammingMode();
  picAddress = 0;
  
  while (!feof(f))
  {
    
    fgets(input, MAX_INPUT, f);


    
    if (0 == strncmp(input, ":00000001FF", 11))
      break;

    
    if (0 == strncmp(input, ":01200700", 9))
    {
      /* Extract config bits, excluding bandgap cals */
      sscanf(input+9, "%XFF", &ivalue);
      configword = (ivalue & 0xFFF) | (configword & 0xF000); 
      continue;
    }

    
    /* Extract number of words to program */
    sscanf(input+1,"%2X", &ivalue);
    /* Sanity check: Is the line long enough? */
    if (ivalue*4+11 > strlen(input))
      continue;

    
    sscanf(input+3, "%2X", &ivalue);
    lineaddress = ivalue << 8;
    sscanf(input+5, "%2X", &ivalue);
    lineaddress += ivalue;

    /* Do a sanity check first */
    if (lineaddress < picAddress)
    {
      /* Stepping back is not allowed! */
      printf("Start address of line %X is below current PIC address %X! Aborting...\n", lineaddress, picAddress);

      setInactiveMode();
      closeLib();
      return ERROR;
    }

    /* Step PIC address forward, if necessary */
    while (lineaddress > picAddress)
    {
      incrementAddress();
      ++picAddress;
    } 

    /* Extract number of words to program */
    sscanf(input+1,"%2X", &ivalue);

    i_ptr = input+9;
    for (word = 0; word < ivalue; ++word)
    {
      /* Extract data word */
      sscanf(i_ptr,"%4X", &data);
      /* Program data word */
      if (ERROR == programAndVerify(data))
      {
        printf("Error while programming memory location %04X: Writing %04X\n", picAddress, data);
        printf("Input line is <%s>\n", input);
        printf("Aborting INHEX16 programming...\n");
        setInactiveMode();
        closeLib();
        return ERROR;
      }
      /* Increment address */
      incrementAddress();
      ++picAddress;

      /* Step to next word */
      i_ptr += 4;
    }


  }

  
  for (; picAddress < 0x3FF; ++picAddress)
    incrementAddress();

  if (ERROR == programAndVerify(osccal))
  {
    printf("Error while programming OSCCAL value %X!\n", osccal);
    setInactiveMode();
    closeLib();
    return ERROR;
  }

  
  loadConfig();

  for ( picAddress = 0x2000; picAddress < 0x2007; ++picAddress)
    incrementAddress();

  if (ERROR == programAndVerify(configword))
  {
    printf("Error while programming CONFIG value %X!\n", configword);
    setInactiveMode();
    closeLib();
    return ERROR;
  }


  setInactiveMode();
  fclose(f);

  return status;
}


